from __future__ import annotations

import typing as t

import pytest
import sqlalchemy as sa
import sqlalchemy.orm as sa_orm
from flask import Flask

from flask_sqlalchemy import SQLAlchemy
from flask_sqlalchemy.session import Session


def test_scope(app: Flask, db: SQLAlchemy) -> None:
    with pytest.raises(RuntimeError):
        db.session()

    with app.app_context():
        first = db.session()
        second = db.session()
        assert first is second
        assert isinstance(first, Session)

    with app.app_context():
        third = db.session()
        assert first is not third


def test_custom_scope(app: Flask, model_class: t.Any) -> None:
    count = 0

    def scope() -> int:
        nonlocal count
        count += 1
        return count

    db = SQLAlchemy(app, model_class=model_class, session_options={"scopefunc": scope})

    with app.app_context():
        first = db.session()
        second = db.session()
        assert first is not second  # a new scope is generated on each call
        first.close()
        second.close()


@pytest.mark.usefixtures("app_ctx")
def test_session_class(app: Flask, model_class: t.Any) -> None:
    class CustomSession(Session):
        pass

    db = SQLAlchemy(
        app, model_class=model_class, session_options={"class_": CustomSession}
    )
    assert isinstance(db.session(), CustomSession)


@pytest.mark.usefixtures("app_ctx")
def test_session_uses_bind_key(app: Flask, model_class: t.Any) -> None:
    app.config["SQLALCHEMY_BINDS"] = {"a": "sqlite://"}
    db = SQLAlchemy(app, model_class=model_class)

    if issubclass(db.Model, (sa_orm.DeclarativeBase, sa_orm.DeclarativeBaseNoMeta)):

        class User(db.Model):
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(sa.Integer, primary_key=True)

        class Post(db.Model):
            __bind_key__ = "a"
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(sa.Integer, primary_key=True)

    else:

        class User(db.Model):  # type: ignore[no-redef]
            id = sa.Column(sa.Integer, primary_key=True)

        class Post(db.Model):  # type: ignore[no-redef]
            __bind_key__ = "a"
            id = sa.Column(sa.Integer, primary_key=True)

    assert db.session.get_bind(mapper=User) is db.engine
    assert db.session.get_bind(mapper=Post) is db.engines["a"]


@pytest.mark.usefixtures("app_ctx")
def test_get_bind_inheritance(app: Flask, model_class: t.Any) -> None:
    app.config["SQLALCHEMY_BINDS"] = {"a": "sqlite://"}
    db = SQLAlchemy(app, model_class=model_class)

    if issubclass(db.Model, (sa_orm.MappedAsDataclass)):

        class User(db.Model):
            __bind_key__ = "a"
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(
                sa.Integer, primary_key=True, init=False
            )
            type: sa_orm.Mapped[str] = sa_orm.mapped_column(
                sa.String, nullable=False, init=False
            )
            __mapper_args__ = {"polymorphic_on": type, "polymorphic_identity": "user"}

        class Admin(User):
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(
                sa.ForeignKey(User.id), primary_key=True, init=False
            )
            org: sa_orm.Mapped[str] = sa_orm.mapped_column(sa.String, nullable=False)
            __mapper_args__ = {"polymorphic_identity": "admin"}

    elif issubclass(db.Model, (sa_orm.DeclarativeBase, sa_orm.DeclarativeBaseNoMeta)):

        class User(db.Model):  # type: ignore[no-redef]
            __bind_key__ = "a"
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(sa.Integer, primary_key=True)
            type: sa_orm.Mapped[str] = sa_orm.mapped_column(sa.String, nullable=False)
            __mapper_args__ = {"polymorphic_on": type, "polymorphic_identity": "user"}

        class Admin(User):  # type: ignore[no-redef]
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(
                sa.ForeignKey(User.id), primary_key=True
            )
            org: sa_orm.Mapped[str] = sa_orm.mapped_column(sa.String, nullable=False)
            __mapper_args__ = {"polymorphic_identity": "admin"}

    else:

        class User(db.Model):  # type: ignore[no-redef]
            __bind_key__ = "a"
            id = sa.Column(sa.Integer, primary_key=True)
            type = sa.Column(sa.String, nullable=False)
            __mapper_args__ = {"polymorphic_on": type, "polymorphic_identity": "user"}

        class Admin(User):  # type: ignore[no-redef]
            id = sa.Column(sa.ForeignKey(User.id), primary_key=True)  # type: ignore[assignment]
            org = sa.Column(sa.String, nullable=False)
            __mapper_args__ = {"polymorphic_identity": "admin"}

    db.create_all()
    db.session.add(Admin(org="pallets"))
    db.session.commit()
    admin = db.session.execute(db.select(Admin)).scalar_one()
    db.session.expire(admin)
    assert admin.org == "pallets"


@pytest.mark.usefixtures("app_ctx")
def test_session_multiple_dbs(app: Flask, model_class: t.Any) -> None:
    app.config["SQLALCHEMY_BINDS"] = {"db1": "sqlite:///"}
    db = SQLAlchemy(app, model_class=model_class)

    if issubclass(db.Model, (sa_orm.MappedAsDataclass)):

        class User(db.Model):
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(
                sa.Integer, primary_key=True, init=False
            )
            name: sa_orm.Mapped[str] = sa_orm.mapped_column(
                sa.String(50), nullable=False, init=False
            )

        class Product(db.Model):
            __bind_key__ = "db1"
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(
                sa.Integer, primary_key=True, init=False
            )
            name: sa_orm.Mapped[str] = sa_orm.mapped_column(
                sa.String(50), nullable=False, init=False
            )

    elif issubclass(db.Model, (sa_orm.DeclarativeBase, sa_orm.DeclarativeBaseNoMeta)):

        class User(db.Model):  # type: ignore[no-redef]
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(sa.Integer, primary_key=True)
            name: sa_orm.Mapped[str] = sa_orm.mapped_column(
                sa.String(50), nullable=False
            )

        class Product(db.Model):  # type: ignore[no-redef]
            __bind_key__ = "db1"
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(sa.Integer, primary_key=True)
            name: sa_orm.Mapped[str] = sa_orm.mapped_column(
                sa.String(50), nullable=False
            )

    else:

        class User(db.Model):  # type: ignore[no-redef]
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.String(50), nullable=False)

        class Product(db.Model):  # type: ignore[no-redef]
            __bind_key__ = "db1"
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.String(50), nullable=False)

    db.create_all()

    db.session.execute(User.__table__.insert(), [{"name": "User1"}, {"name": "User2"}])
    db.session.commit()
    users = db.session.execute(db.select(User)).scalars().all()
    assert len(users) == 2

    db.session.execute(
        Product.__table__.insert(), [{"name": "Product1"}, {"name": "Product2"}]
    )
    db.session.commit()
    products = db.session.execute(db.select(Product)).scalars().all()
    assert len(products) == 2



from __future__ import annotations

import types
import typing as t
from pathlib import Path

import pytest
import sqlalchemy as sa
import sqlalchemy.orm as sa_orm
from flask import Flask
from flask.ctx import AppContext

from flask_sqlalchemy import SQLAlchemy
from flask_sqlalchemy.model import Model


@pytest.fixture
def app(request: pytest.FixtureRequest, tmp_path: Path) -> Flask:
    app = Flask(request.module.__name__, instance_path=str(tmp_path / "instance"))
    app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite://"
    app.config["SQLALCHEMY_RECORD_QUERIES"] = False
    return app


@pytest.fixture
def app_ctx(app: Flask) -> t.Generator[AppContext, None, None]:
    with app.app_context() as ctx:
        yield ctx


# Each test that uses the db or model_class fixture will be tested against
# each of these possible base class setups.
# The first one is None, which will trigger creation of a SQLAlchemy 1.x base.
# The remaining four are used to create SQLAlchemy 2.x bases.
# We defer creation of those classes until the fixture,
# so that each test gets a fresh class with its own metadata.
test_classes = [
    Model,
    (
        "BaseDeclarativeBase",
        (sa_orm.DeclarativeBase,),
        {"metaclass": type(sa_orm.DeclarativeBase)},
    ),
    (
        "BaseDataclassDeclarativeBase",
        (sa_orm.MappedAsDataclass, sa_orm.DeclarativeBase),
        {"metaclass": type(sa_orm.DeclarativeBase)},
    ),
    (
        "BaseDeclarativeBaseNoMeta",
        (sa_orm.DeclarativeBaseNoMeta,),
        {"metaclass": type(sa_orm.DeclarativeBaseNoMeta)},
    ),
    (
        "BaseDataclassDeclarativeBaseNoMeta",
        (
            sa_orm.MappedAsDataclass,
            sa_orm.DeclarativeBaseNoMeta,
        ),
        {"metaclass": type(sa_orm.DeclarativeBaseNoMeta)},
    ),
]


@pytest.fixture(params=test_classes)
def db(app: Flask, request: pytest.FixtureRequest) -> SQLAlchemy:
    if request.param is not Model:
        return SQLAlchemy(app, model_class=types.new_class(*request.param))
    else:
        return SQLAlchemy(app)


@pytest.fixture(params=test_classes)
def model_class(request: pytest.FixtureRequest) -> t.Any:
    if request.param is not Model:
        return types.new_class(*request.param)
    else:
        return request.param


@pytest.fixture
def Todo(app: Flask, db: SQLAlchemy) -> t.Generator[t.Any, None, None]:
    if issubclass(db.Model, (sa_orm.MappedAsDataclass)):

        class Todo(db.Model):
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(
                sa.Integer, init=False, primary_key=True
            )
            title: sa_orm.Mapped[str] = sa_orm.mapped_column(
                sa.String, nullable=True, default=None
            )

    elif issubclass(db.Model, (sa_orm.DeclarativeBase, sa_orm.DeclarativeBaseNoMeta)):

        class Todo(db.Model):  # type: ignore[no-redef]
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(sa.Integer, primary_key=True)
            title: sa_orm.Mapped[str] = sa_orm.mapped_column(sa.String, nullable=True)

    else:

        class Todo(db.Model):  # type: ignore[no-redef]
            id = sa.Column(sa.Integer, primary_key=True)
            title = sa.Column(sa.String)

    with app.app_context():
        db.create_all()

    yield Todo

    with app.app_context():
        db.drop_all()



from __future__ import annotations

import typing as t

import pytest
import sqlalchemy as sa
import sqlalchemy.orm as sa_orm
from flask import Flask
from werkzeug.exceptions import NotFound

from flask_sqlalchemy import SQLAlchemy


@pytest.mark.usefixtures("app_ctx")
def test_view_get_or_404(db: SQLAlchemy, Todo: t.Any) -> None:
    item = Todo()
    db.session.add(item)
    db.session.commit()
    assert db.get_or_404(Todo, 1) is item
    with pytest.raises(NotFound):
        assert db.get_or_404(Todo, 2)


@pytest.mark.usefixtures("app_ctx")
def test_first_or_404(db: SQLAlchemy, Todo: t.Any) -> None:
    db.session.add(Todo(title="a"))
    db.session.commit()
    result = db.first_or_404(db.select(Todo).filter_by(title="a"))
    assert result.title == "a"

    with pytest.raises(NotFound):
        db.first_or_404(db.select(Todo).filter_by(title="b"))


@pytest.mark.usefixtures("app_ctx")
def test_view_one_or_404(db: SQLAlchemy, Todo: t.Any) -> None:
    db.session.add(Todo(title="a"))
    db.session.add(Todo(title="b"))
    db.session.add(Todo(title="b"))
    db.session.commit()
    result = db.one_or_404(db.select(Todo).filter_by(title="a"))
    assert result.title == "a"

    with pytest.raises(NotFound):
        # MultipleResultsFound
        db.one_or_404(db.select(Todo).filter_by(title="b"))

    with pytest.raises(NotFound):
        # NoResultFound
        db.one_or_404(db.select(Todo).filter_by(title="c"))


@pytest.mark.usefixtures("app_ctx")
def test_paginate(db: SQLAlchemy, Todo: t.Any) -> None:
    db.session.add_all(Todo() for _ in range(150))
    db.session.commit()
    p = db.paginate(db.select(Todo))
    assert p.total == 150
    assert len(p.items) == 20
    p2 = p.next()
    assert p2.page == 2
    assert p2.total == 150


# This test creates its own inline model so that it can use that as the type
@pytest.mark.usefixtures("app_ctx")
def test_view_get_or_404_typed(db: SQLAlchemy, app: Flask) -> None:
    # Copied and pasted from conftest.py
    if issubclass(db.Model, (sa_orm.MappedAsDataclass)):

        class Todo(db.Model):
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(
                sa.Integer, init=False, primary_key=True
            )
            title: sa_orm.Mapped[str] = sa_orm.mapped_column(
                sa.String, nullable=True, default=None
            )

    elif issubclass(db.Model, (sa_orm.DeclarativeBase, sa_orm.DeclarativeBaseNoMeta)):

        class Todo(db.Model):  # type: ignore[no-redef]
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(sa.Integer, primary_key=True)
            title: sa_orm.Mapped[str] = sa_orm.mapped_column(sa.String, nullable=True)

    else:

        class Todo(db.Model):  # type: ignore[no-redef]
            id = sa.Column(sa.Integer, primary_key=True)
            title = sa.Column(sa.String)

    db.create_all()

    todo = Todo()
    todo.title = "Python"
    db.session.add(todo)
    db.session.commit()
    result = db.get_or_404(Todo, 1)
    assert result.title == "Python"
    assert result is todo
    if hasattr(t, "assert_type"):
        t.assert_type(result, Todo)
    with pytest.raises(NotFound):
        assert db.get_or_404(Todo, 2)
    db.drop_all()



from __future__ import annotations

import typing as t

import pytest
import sqlalchemy as sa
import sqlalchemy.exc as sa_exc
import sqlalchemy.orm as sa_orm
from flask import Flask

from flask_sqlalchemy import SQLAlchemy
from flask_sqlalchemy.model import DefaultMeta
from flask_sqlalchemy.model import Model


def test_default_metadata(db: SQLAlchemy) -> None:
    assert db.metadata is db.metadatas[None]
    assert db.metadata.info["bind_key"] is None
    assert db.Model.metadata is db.metadata


def test_custom_metadata_1x() -> None:
    metadata = sa.MetaData()
    db = SQLAlchemy(metadata=metadata)
    assert db.metadata is metadata
    assert db.metadata.info["bind_key"] is None
    assert db.Model.metadata is db.metadata


def test_custom_metadata_2x_wrongway() -> None:
    custom_metadata = sa.MetaData()

    class Base(sa_orm.DeclarativeBase):
        pass

    with pytest.deprecated_call():
        db = SQLAlchemy(model_class=Base, metadata=custom_metadata)

        assert db.metadata is Base.metadata
        assert db.metadata.info["bind_key"] is None
        assert db.Model.metadata is db.metadata


def test_custom_metadata_2x() -> None:
    custom_metadata = sa.MetaData()

    class Base(sa_orm.DeclarativeBase):
        metadata = custom_metadata

    db = SQLAlchemy(model_class=Base)

    assert db.metadata is custom_metadata
    assert db.metadata.info["bind_key"] is None
    assert db.Model.metadata is db.metadata


def test_metadata_from_custom_model(model_class: t.Any) -> None:
    if model_class is not Model:
        # In 2.x, SQLAlchemy creates the metadata attribute
        base = model_class
    else:
        # For 1.x, our extension creates the metadata attribute
        base = sa_orm.declarative_base(cls=Model, metaclass=DefaultMeta)
    metadata = base.metadata
    db = SQLAlchemy(model_class=base)
    assert db.Model.metadata is metadata
    assert db.Model.metadata is db.metadata


def test_custom_metadata_overrides_custom_model_legacy() -> None:
    base = sa_orm.declarative_base(cls=Model, metaclass=DefaultMeta)
    metadata = sa.MetaData()
    db = SQLAlchemy(model_class=base, metadata=metadata)
    assert db.Model.metadata is metadata
    assert db.Model.metadata is db.metadata


def test_metadata_per_bind(app: Flask, model_class: t.Any) -> None:
    app.config["SQLALCHEMY_BINDS"] = {"a": "sqlite://"}
    db = SQLAlchemy(app, model_class=model_class)
    assert db.metadatas["a"] is not db.metadata
    assert db.metadatas["a"].info["bind_key"] == "a"


def test_copy_naming_convention(app: Flask, model_class: t.Any) -> None:
    app.config["SQLALCHEMY_BINDS"] = {"a": "sqlite://"}
    if model_class is not Model:
        model_class.metadata = sa.MetaData(
            naming_convention={"pk": "spk_%(table_name)s"}
        )
        db = SQLAlchemy(app, model_class=model_class)
    else:
        db = SQLAlchemy(
            app, metadata=sa.MetaData(naming_convention={"pk": "spk_%(table_name)s"})
        )
    assert db.metadata.naming_convention["pk"] == "spk_%(table_name)s"
    assert db.metadatas["a"].naming_convention == db.metadata.naming_convention


@pytest.mark.usefixtures("app_ctx")
def test_create_drop_all(app: Flask) -> None:
    app.config["SQLALCHEMY_BINDS"] = {"a": "sqlite://"}
    db = SQLAlchemy(app)

    class User(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)

    class Post(db.Model):
        __bind_key__ = "a"
        id = sa.Column(sa.Integer, primary_key=True)

    with pytest.raises(sa_exc.OperationalError):
        db.session.execute(sa.select(User)).scalars()

    with pytest.raises(sa_exc.OperationalError):
        db.session.execute(sa.select(Post)).scalars()

    db.create_all()
    db.session.execute(sa.select(User)).scalars()
    db.session.execute(sa.select(Post)).scalars()
    db.drop_all()

    with pytest.raises(sa_exc.OperationalError):
        db.session.execute(sa.select(User)).scalars()

    with pytest.raises(sa_exc.OperationalError):
        db.session.execute(sa.select(Post)).scalars()


@pytest.mark.usefixtures("app_ctx")
@pytest.mark.parametrize("bind_key", ["a", ["a"]])
def test_create_key_spec(app: Flask, bind_key: str | list[str | None]) -> None:
    app.config["SQLALCHEMY_BINDS"] = {"a": "sqlite://"}
    db = SQLAlchemy(app)

    class User(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)

    class Post(db.Model):
        __bind_key__ = "a"
        id = sa.Column(sa.Integer, primary_key=True)

    db.create_all(bind_key=bind_key)
    db.session.execute(sa.select(Post)).scalars()

    with pytest.raises(sa_exc.OperationalError):
        db.session.execute(sa.select(User)).scalars()


@pytest.mark.usefixtures("app_ctx")
def test_reflect(app: Flask) -> None:
    app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///user.db"
    app.config["SQLALCHEMY_BINDS"] = {"post": "sqlite:///post.db"}
    db = SQLAlchemy(app)
    db.Table("user", sa.Column("id", sa.Integer, primary_key=True))
    db.Table("post", sa.Column("id", sa.Integer, primary_key=True), bind_key="post")
    db.create_all()

    del app.extensions["sqlalchemy"]
    db = SQLAlchemy(app)
    assert not db.metadata.tables
    db.reflect()
    assert "user" in db.metadata.tables
    assert "post" in db.metadatas["post"].tables



from __future__ import annotations

import os

import pytest
import sqlalchemy as sa
import sqlalchemy.orm as sa_orm
from flask import Flask

from flask_sqlalchemy import SQLAlchemy
from flask_sqlalchemy.record_queries import get_recorded_queries


@pytest.mark.usefixtures("app_ctx")
def test_query_info(app: Flask) -> None:
    app.config["SQLALCHEMY_RECORD_QUERIES"] = True
    db = SQLAlchemy(app)

    # Copied and pasted from conftest.py
    if issubclass(db.Model, (sa_orm.MappedAsDataclass)):

        class Todo(db.Model):
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(
                sa.Integer, init=False, primary_key=True
            )
            title: sa_orm.Mapped[str] = sa_orm.mapped_column(
                sa.String, nullable=True, default=None
            )

    elif issubclass(db.Model, (sa_orm.DeclarativeBase, sa_orm.DeclarativeBaseNoMeta)):

        class Todo(db.Model):  # type: ignore[no-redef]
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(sa.Integer, primary_key=True)
            title: sa_orm.Mapped[str] = sa_orm.mapped_column(sa.String, nullable=True)

    else:

        class Todo(db.Model):  # type: ignore[no-redef]
            id = sa.Column(sa.Integer, primary_key=True)
            title = sa.Column(sa.String)

    db.create_all()
    db.session.execute(sa.select(Todo).filter(Todo.id < 5)).scalars()
    info = get_recorded_queries()[-1]
    assert info.statement is not None
    assert "SELECT" in info.statement
    assert "FROM todo" in info.statement
    assert info.parameters[0][0] == 5
    assert info.duration == info.end_time - info.start_time
    assert os.path.join("tests", "test_record_queries.py:") in info.location
    assert "(test_query_info)" in info.location



from __future__ import annotations

import typing as t

import pytest
import sqlalchemy as sa
import sqlalchemy.orm as sa_orm
from flask import Flask
from werkzeug.exceptions import NotFound

from flask_sqlalchemy import SQLAlchemy
from flask_sqlalchemy.record_queries import get_recorded_queries


@pytest.mark.usefixtures("app_ctx")
def test_get_or_404(db: SQLAlchemy, Todo: t.Any) -> None:
    item = Todo()
    db.session.add(item)
    db.session.commit()
    assert db.get_or_404(Todo, 1) is item

    with pytest.raises(NotFound):
        db.get_or_404(Todo, 2)


def test_get_or_404_kwargs(app: Flask, model_class: t.Any) -> None:
    app.config["SQLALCHEMY_RECORD_QUERIES"] = True
    db = SQLAlchemy(app, model_class=model_class)

    if issubclass(db.Model, (sa_orm.MappedAsDataclass)):

        class User(db.Model):
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(
                sa.Integer, primary_key=True, init=False
            )

        class Todo(db.Model):
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(
                sa.Integer, primary_key=True, init=False
            )
            user_id: sa_orm.Mapped[int] = sa_orm.mapped_column(
                sa.ForeignKey(User.id), init=False
            )
            user: sa_orm.Mapped[User] = sa_orm.relationship(User)

    elif issubclass(db.Model, (sa_orm.DeclarativeBase, sa_orm.DeclarativeBaseNoMeta)):

        class User(db.Model):  # type: ignore[no-redef]
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(sa.Integer, primary_key=True)

        class Todo(db.Model):  # type: ignore[no-redef]
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(sa.Integer, primary_key=True)
            user_id: sa_orm.Mapped[int] = sa_orm.mapped_column(sa.ForeignKey(User.id))
            user: sa_orm.Mapped[User] = sa_orm.relationship(User)

    else:

        class User(db.Model):  # type: ignore[no-redef]
            id = sa.Column(db.Integer, primary_key=True)  # type: ignore[var-annotated]

        class Todo(db.Model):  # type: ignore[no-redef]
            id = sa.Column(sa.Integer, primary_key=True)
            user_id = sa.Column(sa.ForeignKey(User.id))  # type: ignore[var-annotated]
            user = db.relationship(User)

    with app.app_context():
        db.create_all()
        db.session.add(Todo(user=User()))
        db.session.commit()

    with app.app_context():
        item = db.get_or_404(Todo, 1, options=[db.joinedload(Todo.user)])
        assert item.user.id == 1
        # one query with join, no second query when accessing relationship
        assert len(get_recorded_queries()) == 1



from __future__ import annotations

import typing as t
from datetime import datetime

import pytest
import sqlalchemy as sa
import sqlalchemy.exc as sa_exc
import sqlalchemy.orm as sa_orm
from flask import Flask

from flask_sqlalchemy import SQLAlchemy
from flask_sqlalchemy.model import DefaultMeta
from flask_sqlalchemy.model import Model


def test_default_model_class_1x(app: Flask) -> None:
    db = SQLAlchemy(app)

    assert db.Model.query_class is db.Query
    assert db.Model.metadata is db.metadata
    assert issubclass(db.Model, Model)
    assert isinstance(db.Model, DefaultMeta)


def test_custom_model_class_1x(app: Flask) -> None:
    class CustomModel(Model):
        pass

    db = SQLAlchemy(app, model_class=CustomModel)
    assert issubclass(db.Model, CustomModel)
    assert isinstance(db.Model, DefaultMeta)


@pytest.mark.usefixtures("app_ctx")
@pytest.mark.parametrize("base", [Model, object])
def test_custom_declarative_class_1x(app: Flask, base: t.Any) -> None:
    class CustomMeta(DefaultMeta):
        pass

    CustomModel = sa_orm.declarative_base(cls=base, name="Model", metaclass=CustomMeta)
    db = SQLAlchemy(app, model_class=CustomModel)
    assert db.Model is CustomModel
    assert db.Model.query_class is db.Query
    assert "query" in db.Model.__dict__


def test_declarativebase_2x(app: Flask) -> None:
    class Base(sa_orm.DeclarativeBase):
        pass

    db = SQLAlchemy(app, model_class=Base)
    assert issubclass(db.Model, sa_orm.DeclarativeBase)
    assert isinstance(db.Model, sa_orm.decl_api.DeclarativeAttributeIntercept)


def test_declarativebasenometa_2x(app: Flask) -> None:
    class Base(sa_orm.DeclarativeBaseNoMeta):
        pass

    db = SQLAlchemy(app, model_class=Base)
    assert issubclass(db.Model, sa_orm.DeclarativeBaseNoMeta)
    assert not isinstance(db.Model, sa_orm.decl_api.DeclarativeAttributeIntercept)


def test_declarativebasemapped_2x(app: Flask) -> None:
    class Base(sa_orm.DeclarativeBase, sa_orm.MappedAsDataclass):
        pass

    db = SQLAlchemy(app, model_class=Base)
    assert issubclass(db.Model, sa_orm.DeclarativeBase)
    assert isinstance(db.Model, sa_orm.decl_api.DCTransformDeclarative)


def test_declarativebasenometamapped_2x(app: Flask) -> None:
    class Base(sa_orm.DeclarativeBaseNoMeta, sa_orm.MappedAsDataclass):
        pass

    db = SQLAlchemy(app, model_class=Base)
    assert issubclass(db.Model, sa_orm.DeclarativeBaseNoMeta)
    assert isinstance(db.Model, sa_orm.decl_api.DCTransformDeclarative)


@pytest.mark.usefixtures("app_ctx")
def test_declaredattr(app: Flask, model_class: t.Any) -> None:
    if model_class is Model:

        class IdModel(Model):
            @sa.orm.declared_attr
            @classmethod
            def id(cls: type[Model]):  # type: ignore[no-untyped-def]
                for base in cls.__mro__[1:-1]:
                    if getattr(base, "__table__", None) is not None and hasattr(
                        base, "id"
                    ):
                        return sa.Column(sa.ForeignKey(base.id), primary_key=True)
                return sa.Column(sa.Integer, primary_key=True)

        db = SQLAlchemy(app, model_class=IdModel)

        class User(db.Model):
            name = db.Column(db.String)

        class Employee(User):
            title = db.Column(db.String)

    else:

        class Base(sa_orm.DeclarativeBase):
            @sa_orm.declared_attr
            @classmethod
            def id(cls: type[sa_orm.DeclarativeBase]) -> sa_orm.Mapped[int]:
                for base in cls.__mro__[1:-1]:
                    if getattr(base, "__table__", None) is not None and hasattr(
                        base, "id"
                    ):
                        return sa_orm.mapped_column(
                            db.ForeignKey(base.id), primary_key=True
                        )
                return sa_orm.mapped_column(db.Integer, primary_key=True)

        db = SQLAlchemy(app, model_class=Base)

        class User(db.Model):  # type: ignore[no-redef]
            name: sa_orm.Mapped[str] = sa_orm.mapped_column(db.String)

        class Employee(User):  # type: ignore[no-redef]
            title: sa_orm.Mapped[str] = sa_orm.mapped_column(db.String)

    db.create_all()
    db.session.add(Employee(name="Emp Loyee", title="Admin"))
    db.session.commit()
    user = db.session.execute(db.select(User)).scalar()
    employee = db.session.execute(db.select(Employee)).scalar()
    assert user is not None
    assert employee is not None
    assert user.id == 1
    assert employee.id == 1


@pytest.mark.usefixtures("app_ctx")
def test_abstractmodel(app: Flask, model_class: t.Any) -> None:
    db = SQLAlchemy(app, model_class=model_class)

    if issubclass(db.Model, (sa_orm.MappedAsDataclass)):

        class TimestampModel(db.Model):
            __abstract__ = True
            created: sa_orm.Mapped[datetime] = sa_orm.mapped_column(
                db.DateTime, nullable=False, insert_default=datetime.utcnow, init=False
            )
            updated: sa_orm.Mapped[datetime] = sa_orm.mapped_column(
                db.DateTime,
                insert_default=datetime.utcnow,
                onupdate=datetime.utcnow,
                init=False,
            )

        class Post(TimestampModel):
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(
                db.Integer, primary_key=True, init=False
            )
            title: sa_orm.Mapped[str] = sa_orm.mapped_column(db.String, nullable=False)

    elif issubclass(db.Model, (sa_orm.DeclarativeBase, sa_orm.DeclarativeBaseNoMeta)):

        class TimestampModel(db.Model):  # type: ignore[no-redef]
            __abstract__ = True
            created: sa_orm.Mapped[datetime] = sa_orm.mapped_column(
                db.DateTime, nullable=False, default=datetime.utcnow
            )
            updated: sa_orm.Mapped[datetime] = sa_orm.mapped_column(
                db.DateTime, default=datetime.utcnow, onupdate=datetime.utcnow
            )

        class Post(TimestampModel):  # type: ignore[no-redef]
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(db.Integer, primary_key=True)
            title: sa_orm.Mapped[str] = sa_orm.mapped_column(db.String, nullable=False)

    else:

        class TimestampModel(db.Model):  # type: ignore[no-redef]
            __abstract__ = True
            created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
            updated = db.Column(
                db.DateTime, onupdate=datetime.utcnow, default=datetime.utcnow
            )

        class Post(TimestampModel):  # type: ignore[no-redef]
            id = db.Column(db.Integer, primary_key=True)
            title = db.Column(db.String, nullable=False)

    db.create_all()
    db.session.add(Post(title="Admin Post"))
    db.session.commit()
    post = db.session.execute(db.select(Post)).scalar()
    assert post is not None
    assert post.created is not None
    assert post.updated is not None


@pytest.mark.usefixtures("app_ctx")
def test_mixinmodel(app: Flask, model_class: t.Any) -> None:
    db = SQLAlchemy(app, model_class=model_class)

    if issubclass(db.Model, (sa_orm.MappedAsDataclass)):

        class TimestampMixin(sa_orm.MappedAsDataclass):
            created: sa_orm.Mapped[datetime] = sa_orm.mapped_column(
                db.DateTime, nullable=False, insert_default=datetime.utcnow, init=False
            )
            updated: sa_orm.Mapped[datetime] = sa_orm.mapped_column(
                db.DateTime,
                insert_default=datetime.utcnow,
                onupdate=datetime.utcnow,
                init=False,
            )

        class Post(TimestampMixin, db.Model):
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(
                db.Integer, primary_key=True, init=False
            )
            title: sa_orm.Mapped[str] = sa_orm.mapped_column(db.String, nullable=False)

    elif issubclass(db.Model, (sa_orm.DeclarativeBase, sa_orm.DeclarativeBaseNoMeta)):

        class TimestampMixin:  # type: ignore[no-redef]
            created: sa_orm.Mapped[datetime] = sa_orm.mapped_column(
                db.DateTime, nullable=False, default=datetime.utcnow
            )
            updated: sa_orm.Mapped[datetime] = sa_orm.mapped_column(
                db.DateTime, default=datetime.utcnow, onupdate=datetime.utcnow
            )

        class Post(TimestampMixin, db.Model):  # type: ignore[no-redef]
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(db.Integer, primary_key=True)
            title: sa_orm.Mapped[str] = sa_orm.mapped_column(db.String, nullable=False)

    else:

        class TimestampMixin:  # type: ignore[no-redef]
            created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
            updated = db.Column(
                db.DateTime, onupdate=datetime.utcnow, default=datetime.utcnow
            )

        class Post(TimestampMixin, db.Model):  # type: ignore[no-redef]
            id = db.Column(db.Integer, primary_key=True)
            title = db.Column(db.String, nullable=False)

    db.create_all()
    db.session.add(Post(title="Admin Post"))
    db.session.commit()
    post = db.session.execute(db.select(Post)).scalar()
    assert post is not None
    assert post.created is not None
    assert post.updated is not None


@pytest.mark.usefixtures("app_ctx")
def test_model_repr(db: SQLAlchemy) -> None:
    class User(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)

    db.create_all()
    user = User()

    if issubclass(db.Model, sa_orm.MappedAsDataclass):
        assert repr(user) == "test_model_repr.<locals>.User()"
    else:
        assert repr(user) == f"<User (transient {id(user)})>"
        db.session.add(user)
        assert repr(user) == f"<User (pending {id(user)})>"
        db.session.flush()
        assert repr(user) == f"<User {user.id}>"


@pytest.mark.usefixtures("app_ctx")
def test_too_many_bases(app: Flask) -> None:
    class Base(sa_orm.DeclarativeBase, sa_orm.DeclarativeBaseNoMeta):  # type: ignore[misc]
        pass

    with pytest.raises(ValueError):
        SQLAlchemy(app, model_class=Base)


@pytest.mark.usefixtures("app_ctx")
def test_disable_autonaming_true_sql1(app: Flask) -> None:
    db = SQLAlchemy(app, disable_autonaming=True)

    with pytest.raises(sa_exc.InvalidRequestError):

        class User(db.Model):
            id = sa.Column(sa.Integer, primary_key=True)


@pytest.mark.usefixtures("app_ctx")
def test_disable_autonaming_true_sql2(app: Flask) -> None:
    class Base(sa_orm.DeclarativeBase):
        pass

    db = SQLAlchemy(app, model_class=Base, disable_autonaming=True)

    with pytest.raises(sa_exc.InvalidRequestError):

        class User(db.Model):
            id: sa_orm.Mapped[int] = sa_orm.mapped_column(sa.Integer, primary_key=True)



from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy import ForeignKey
from datetime import datetime
from datetime import timezone

from flask import url_for

from flaskr import db
from flaskr.auth.models import User


def now_utc() -> datetime:
    return datetime.now(timezone.utc)


class Post(db.Model):
    id: Mapped[int] = mapped_column(primary_key=True)
    author_id: Mapped[int] = mapped_column(ForeignKey("user.id"))
    created: Mapped[datetime] = mapped_column(default=now_utc)
    title: Mapped[str]
    body: Mapped[str]

    # User object backed by author_id
    # lazy="joined" means the user is returned with the post in one query
    author: Mapped[User] = relationship(lazy="joined", back_populates="posts")

    @property
    def update_url(self) -> str:
        return url_for("blog.update", id=self.id)

    @property
    def delete_url(self) -> str:
        return url_for("blog.delete", id=self.id)



from datetime import datetime
from datetime import timezone

from flask import flash
from flask import Flask
from flask import redirect
from flask import render_template
from flask import request
from flask import url_for
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped, mapped_column


app = Flask(__name__)
app.secret_key = "Achee6phIexoh8dagiQuew0ephuga4Ih"
app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///todo.sqlite"


class Base(DeclarativeBase):
    pass


db = SQLAlchemy(app, model_class=Base)


def now_utc() -> datetime:
    return datetime.now(timezone.utc)


class Todo(db.Model):
    id: Mapped[int] = mapped_column(primary_key=True)
    title: Mapped[str]
    text: Mapped[str]
    done: Mapped[bool] = mapped_column(default=False)
    pub_date: Mapped[datetime] = mapped_column(default=now_utc)


with app.app_context():
    db.create_all()


@app.route("/")
def show_all():
    select = db.select(Todo).order_by(Todo.pub_date.desc())
    todos = db.session.execute(select).scalars()
    return render_template("show_all.html", todos=todos)


@app.route("/new", methods=["GET", "POST"])
def new():
    if request.method == "POST":
        if not request.form["title"]:
            flash("Title is required", "error")
        elif not request.form["text"]:
            flash("Text is required", "error")
        else:
            todo = Todo(title=request.form["title"], text=request.form["text"])
            db.session.add(todo)
            db.session.commit()
            flash("Todo item was successfully created")
            return redirect(url_for("show_all"))

    return render_template("new.html")


@app.route("/update", methods=["POST"])
def update_done():
    for todo in db.session.execute(db.select(Todo)).scalars():
        todo.done = f"done.{todo.id}" in request.form

    flash("Updated status")
    db.session.commit()
    return redirect(url_for("show_all"))



from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy.orm import Mapped, mapped_column, relationship
from werkzeug.security import check_password_hash
from werkzeug.security import generate_password_hash

from flaskr import db

if TYPE_CHECKING:
    from ..blog.models import Post


class User(db.Model):
    id: Mapped[int] = mapped_column(primary_key=True)
    username: Mapped[str] = mapped_column(unique=True)
    password_hash: Mapped[str]
    posts: Mapped[list[Post]] = relationship("Post", back_populates="author")

    def set_password(self, value: str) -> None:
        """Store the password as a hash for security."""
        self.password_hash = generate_password_hash(value)

    # allow password = "..." to set a password
    password = property(fset=set_password)

    def check_password(self, value: str) -> bool:
        return check_password_hash(self.password_hash, value)
